@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2002-2012 Free Software Foundation, Inc.
+ * Copyright (C) 2002-2013 Free Software Foundation, Inc.
  *
  * Author: Nikos Mavrogiannopoulos
  *
@@ -7,7 +7,7 @@
  *
  * The GnuTLS is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 3 of
+ * as published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful, but
@@ -45,6 +45,7 @@
 #include <gnutls_rsa_export.h>
 #include <gnutls_extensions.h>
 #include <system.h>
+#include <random.h>
 #include <gnutls/dtls.h>
 #include <timespec.h>
 
@@ -229,11 +230,15 @@ _gnutls_session_cert_type_supported (gnutls_session_t session,
 inline static void
 deinit_internal_params (gnutls_session_t session)
 {
+#if defined(ENABLE_DHE) || defined(ENABLE_ANON)
   if (session->internals.params.free_dh_params)
     gnutls_dh_params_deinit (session->internals.params.dh_params);
+#endif
 
+#ifdef ENABLE_RSA_EXPORT
   if (session->internals.params.free_rsa_params)
     gnutls_rsa_params_deinit (session->internals.params.rsa_params);
+#endif
 
   _gnutls_handshake_hash_buffers_clear (session);
 
@@ -264,7 +269,7 @@ _gnutls_handshake_internal_state_init (gnutls_session_t session)
   
   session->internals.dtls.hsk_read_seq = 0;
   session->internals.dtls.hsk_write_seq = 0;
-  gettime(&session->internals.dtls.handshake_start_time);
+  session->internals.handshake_large_loops = 0;
 }
 
 void
@@ -279,7 +284,6 @@ _gnutls_handshake_internal_state_clear (gnutls_session_t session)
   session->internals.handshake_endtime = 0;
 }
 
-#define MIN_DH_BITS 727
 /**
  * gnutls_init:
  * @session: is a pointer to a #gnutls_session_t structure.
@@ -333,6 +337,7 @@ gnutls_init (gnutls_session_t * session, unsigned int flags)
   _gnutls_buffer_init (&(*session)->internals.handshake_hash_buffer);
   _gnutls_buffer_init (&(*session)->internals.hb_remote_data);
   _gnutls_buffer_init (&(*session)->internals.hb_local_data);
+  _gnutls_buffer_init (&(*session)->internals.record_presend_buffer);
 
   _mbuffer_head_init (&(*session)->internals.record_buffer);
   _mbuffer_head_init (&(*session)->internals.record_send_buffer);
@@ -343,8 +348,6 @@ gnutls_init (gnutls_session_t * session, unsigned int flags)
 
   (*session)->internals.expire_time = DEFAULT_EXPIRE_TIME;      /* one hour default */
 
-  gnutls_dh_set_prime_bits ((*session), MIN_DH_BITS);
-
   gnutls_handshake_set_max_packet_length ((*session),
                                           MAX_HANDSHAKE_PACKET_SIZE);
 
@@ -434,6 +437,8 @@ gnutls_deinit (gnutls_session_t session)
   if (session == NULL)
     return;
 
+  _gnutls_rnd_refresh();
+
   /* remove auth info firstly */
   _gnutls_free_auth_info (session);
 
@@ -451,6 +456,7 @@ gnutls_deinit (gnutls_session_t session)
   _gnutls_buffer_clear (&session->internals.handshake_hash_buffer);
   _gnutls_buffer_clear (&session->internals.hb_remote_data);
   _gnutls_buffer_clear (&session->internals.hb_local_data);
+  _gnutls_buffer_clear (&session->internals.record_presend_buffer);
 
   _mbuffer_head_clear (&session->internals.record_buffer);
   _mbuffer_head_clear (&session->internals.record_recv_buffer);
@@ -481,7 +487,6 @@ gnutls_deinit (gnutls_session_t session)
 
   _gnutls_mpi_release (&session->key.dh_secret);
 
-  memset (session, 0, sizeof (struct gnutls_session_int));
   gnutls_free (session);
 }
 
@@ -1417,3 +1422,44 @@ gnutls_session_get_random (gnutls_session_t session, gnutls_datum_t* client, gnu
       server->size = sizeof(session->security_parameters.server_random);
     }
 }
+
+unsigned int
+timespec_sub_ms (struct timespec *a, struct timespec *b)
+{
+  return (a->tv_sec * 1000 + a->tv_nsec / (1000 * 1000) -
+          (b->tv_sec * 1000 + b->tv_nsec / (1000 * 1000)));
+}
+
+/**
+ * gnutls_handshake_set_random:
+ * @session: is a #gnutls_session_t structure.
+ * @random: a random value of 32-bytes
+ *
+ * This function will explicitly set the server or client hello 
+ * random value in the subsequent TLS handshake. The random value 
+ * should be a 32-byte value.
+ *
+ * Note that this function should not normally be used as gnutls
+ * will select automatically a random value for the handshake.
+ *
+ * This function should not be used when resuming a session.
+ *
+ * Returns: %GNUTLS_E_SUCCESS on success, or an error code.
+ *
+ * Since 3.1.9
+ **/
+int
+gnutls_handshake_set_random (gnutls_session_t session, const gnutls_datum_t* random)
+{
+  if (random->size != GNUTLS_RANDOM_SIZE)
+    return GNUTLS_E_INVALID_REQUEST;
+  
+  session->internals.sc_random_set = 1;
+  if (session->security_parameters.entity == GNUTLS_CLIENT)
+    memcpy(session->internals.resumed_security_parameters.client_random, random->data, random->size);
+  else
+    memcpy(session->internals.resumed_security_parameters.server_random, random->data, random->size);
+
+  return 0;
+}
+
