@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2003-2012 Free Software Foundation, Inc.
+ * Copyright (C) 2012 Nikos Mavrogiannopoulos
  *
  * Author: Nikos Mavrogiannopoulos
  *
@@ -7,7 +8,7 @@
  *
  * The GnuTLS is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 3 of
+ * as published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful, but
@@ -68,7 +69,7 @@ _decode_pkcs12_auth_safe (ASN1_TYPE pkcs12, ASN1_TYPE * authen_safe,
    */
 
   result =
-    _gnutls_x509_read_value (pkcs12, "authSafe.content", &auth_safe, 1);
+    _gnutls_x509_read_string (pkcs12, "authSafe.content", &auth_safe, ASN1_ETYPE_OCTET_STRING);
   if (result < 0)
     {
       gnutls_assert ();
@@ -276,6 +277,37 @@ gnutls_pkcs12_export (gnutls_pkcs12_t pkcs12,
                                   output_data, output_data_size);
 }
 
+/**
+ * gnutls_pkcs12_export2:
+ * @pkcs12: Holds the pkcs12 structure
+ * @format: the format of output params. One of PEM or DER.
+ * @out: will contain a structure PEM or DER encoded
+ *
+ * This function will export the pkcs12 structure to DER or PEM format.
+ *
+ * The output buffer is allocated using gnutls_malloc().
+ *
+ * If the structure is PEM encoded, it will have a header
+ * of "BEGIN PKCS12".
+ *
+ * Returns: In case of failure a negative error code will be
+ *   returned, and 0 on success.
+ *
+ * Since: 3.1.3
+ **/
+int
+gnutls_pkcs12_export2 (gnutls_pkcs12_t pkcs12,
+                      gnutls_x509_crt_fmt_t format, gnutls_datum_t *out)
+{
+  if (pkcs12 == NULL)
+    {
+      gnutls_assert ();
+      return GNUTLS_E_INVALID_REQUEST;
+    }
+
+  return _gnutls_x509_export_int2 (pkcs12->pkcs12, format, PEM_PKCS12, out);
+}
+
 static int
 oid2bag (const char *oid)
 {
@@ -312,24 +344,6 @@ bag_to_oid (int bag)
   return NULL;
 }
 
-static inline char *
-ucs2_to_ascii (char *data, int size)
-{
-  int i, j;
-
-  for (i = 0; i < size / 2; i++)
-    {
-      j = 2 * i + 1;
-      if (isascii (data[j]))
-        data[i] = data[i * 2 + 1];
-      else
-        data[i] = '?';
-    }
-  data[i] = 0;
-
-  return data;
-}
-
 /* Decodes the SafeContents, and puts the output in
  * the given bag. 
  */
@@ -342,8 +356,8 @@ _pkcs12_decode_safe_contents (const gnutls_datum_t * content,
   int len, result;
   int bag_type;
   gnutls_datum_t attr_val;
+  gnutls_datum_t t;
   int count = 0, i, attributes, j;
-  size_t size;
 
   /* Step 1. Extract the SEQUENCE.
    */
@@ -406,7 +420,7 @@ _pkcs12_decode_safe_contents (const gnutls_datum_t * content,
 
       snprintf (root, sizeof (root), "?%u.bagValue", i + 1);
 
-      result = _gnutls_x509_read_value (c2, root, &bag->element[i].data, 0);
+      result = _gnutls_x509_read_value (c2, root, &bag->element[i].data);
       if (result < 0)
         {
           gnutls_assert ();
@@ -464,40 +478,41 @@ _pkcs12_decode_safe_contents (const gnutls_datum_t * content,
 
             if (strcmp (oid, KEY_ID_OID) == 0)
               {
-                size = attr_val.size;
-
                 result =
-                  _gnutls_x509_decode_octet_string (NULL, attr_val.data, size,
-                                                    attr_val.data, &size);
-                attr_val.size = size;
+                  _gnutls_x509_decode_string (ASN1_ETYPE_OCTET_STRING, attr_val.data, 
+                                              attr_val.size, &t);
+                _gnutls_free_datum (&attr_val);
                 if (result < 0)
                   {
-                    _gnutls_free_datum (&attr_val);
                     gnutls_assert ();
                     _gnutls_debug_log
                       ("Error decoding PKCS12 Bag Attribute OID '%s'\n", oid);
                     continue;
                   }
+                
+                attr_val.data = t.data;
+                attr_val.size = t.size;
+                
                 bag->element[i].local_key_id = attr_val;
               }
             else if (strcmp (oid, FRIENDLY_NAME_OID) == 0)
               {
-                size = attr_val.size;
                 result =
-                  _gnutls_x509_decode_octet_string ("BMPString",
-                                                    attr_val.data, size,
-                                                    attr_val.data, &size);
-                attr_val.size = size;
+                  _gnutls_x509_decode_string (ASN1_ETYPE_BMP_STRING,
+                                              attr_val.data, attr_val.size, &t);
+                _gnutls_free_datum (&attr_val);
                 if (result < 0)
                   {
-                    _gnutls_free_datum (&attr_val);
                     gnutls_assert ();
                     _gnutls_debug_log
                       ("Error decoding PKCS12 Bag Attribute OID '%s'\n", oid);
                     continue;
                   }
-                bag->element[i].friendly_name =
-                  ucs2_to_ascii ((char*)attr_val.data, attr_val.size);
+
+                attr_val.data = t.data;
+                attr_val.size = t.size;
+
+                bag->element[i].friendly_name = (char*)t.data;
               }
             else
               {
@@ -535,7 +550,7 @@ _parse_safe_contents (ASN1_TYPE sc, const char *sc_name,
   /* Step 1. Extract the content.
    */
 
-  result = _gnutls_x509_read_value (sc, sc_name, &content, 1);
+  result = _gnutls_x509_read_string (sc, sc_name, &content, ASN1_ETYPE_OCTET_STRING);
   if (result < 0)
     {
       gnutls_assert ();
@@ -634,7 +649,7 @@ gnutls_pkcs12_get_bag (gnutls_pkcs12_t pkcs12,
   bag->element[0].type = GNUTLS_BAG_ENCRYPTED;
   bag->bag_elements = 1;
 
-  result = _gnutls_x509_read_value (c2, root2, &bag->element[0].data, 0);
+  result = _gnutls_x509_read_value (c2, root2, &bag->element[0].data);
   if (result < 0)
     {
       gnutls_assert ();
@@ -903,7 +918,7 @@ gnutls_pkcs12_generate_mac (gnutls_pkcs12_t pkcs12, const char *pass)
 
   /* Generate the key.
    */
-  result = _gnutls_pkcs12_string_to_key (3 /*MAC*/, salt, sizeof (salt),
+  result = _gnutls_pkcs12_string_to_key (GNUTLS_MAC_SHA1, 3 /*MAC*/, salt, sizeof (salt),
                                          iter, pass, sizeof (key), key);
   if (result < 0)
     {
@@ -986,7 +1001,8 @@ cleanup:
 int
 gnutls_pkcs12_verify_mac (gnutls_pkcs12_t pkcs12, const char *pass)
 {
-  uint8_t key[20];
+  uint8_t key[MAX_HASH_SIZE];
+  char oid[MAX_OID_SIZE];
   int result;
   unsigned int iter;
   int len;
@@ -994,8 +1010,11 @@ gnutls_pkcs12_verify_mac (gnutls_pkcs12_t pkcs12, const char *pass)
   gnutls_datum_t tmp = { NULL, 0 }, salt =
   {
   NULL, 0};
-  uint8_t sha_mac[20];
-  uint8_t sha_mac_orig[20];
+  uint8_t mac_output[MAX_HASH_SIZE];
+  uint8_t mac_output_orig[MAX_HASH_SIZE];
+  gnutls_mac_algorithm_t algo;
+  unsigned mac_len, key_len;
+  
 
   if (pkcs12 == NULL)
     {
@@ -1013,22 +1032,40 @@ gnutls_pkcs12_verify_mac (gnutls_pkcs12_t pkcs12, const char *pass)
       iter = 1;                 /* the default */
     }
 
+  len = sizeof(oid);
+  result =
+    asn1_read_value(pkcs12->pkcs12, "macData.mac.digestAlgorithm.algorithm",
+                    oid, &len);
+  if (result != ASN1_SUCCESS)
+    {
+      gnutls_assert();
+      return _gnutls_asn2err(result);
+    }
+
+  algo = _gnutls_x509_oid_to_mac(oid);
+  if (algo == GNUTLS_MAC_UNKNOWN)
+    {
+      gnutls_assert();
+      return GNUTLS_E_UNKNOWN_HASH_ALGORITHM;
+    }
+
+  mac_len = _gnutls_hmac_get_algo_len(algo);
+  key_len = mac_len;
 
   /* Read the salt from the structure.
    */
   result =
-    _gnutls_x509_read_value (pkcs12->pkcs12, "macData.macSalt", &salt, 0);
-  if (result != ASN1_SUCCESS)
+    _gnutls_x509_read_value (pkcs12->pkcs12, "macData.macSalt", &salt);
+  if (result < 0)
     {
       gnutls_assert ();
-      result = _gnutls_asn2err (result);
       goto cleanup;
     }
 
   /* Generate the key.
    */
-  result = _gnutls_pkcs12_string_to_key (3 /*MAC*/, salt.data, salt.size,
-                                         iter, pass, sizeof (key), key);
+  result = _gnutls_pkcs12_string_to_key (algo, 3 /*MAC*/, salt.data, salt.size,
+                                         iter, pass, key_len, key);
   if (result < 0)
     {
       gnutls_assert ();
@@ -1048,7 +1085,7 @@ gnutls_pkcs12_verify_mac (gnutls_pkcs12_t pkcs12, const char *pass)
 
   /* MAC the data
    */
-  result = _gnutls_hmac_init (&td1, GNUTLS_MAC_SHA1, key, sizeof (key));
+  result = _gnutls_hmac_init (&td1, algo, key, key_len);
   if (result < 0)
     {
       gnutls_assert ();
@@ -1058,11 +1095,11 @@ gnutls_pkcs12_verify_mac (gnutls_pkcs12_t pkcs12, const char *pass)
   _gnutls_hmac (&td1, tmp.data, tmp.size);
   _gnutls_free_datum (&tmp);
 
-  _gnutls_hmac_deinit (&td1, sha_mac);
+  _gnutls_hmac_deinit (&td1, mac_output);
 
-  len = sizeof (sha_mac_orig);
+  len = sizeof (mac_output_orig);
   result =
-    asn1_read_value (pkcs12->pkcs12, "macData.mac.digest", sha_mac_orig,
+    asn1_read_value (pkcs12->pkcs12, "macData.mac.digest", mac_output_orig,
                      &len);
   if (result != ASN1_SUCCESS)
     {
@@ -1071,7 +1108,8 @@ gnutls_pkcs12_verify_mac (gnutls_pkcs12_t pkcs12, const char *pass)
       goto cleanup;
     }
 
-  if (memcmp (sha_mac_orig, sha_mac, sizeof (sha_mac)) != 0)
+  if ((unsigned)len != mac_len ||
+      memcmp (mac_output_orig, mac_output, len) != 0)
     {
       gnutls_assert ();
       return GNUTLS_E_MAC_VERIFY_FAILED;
@@ -1196,7 +1234,7 @@ write_attributes (gnutls_pkcs12_bag_t bag, int elem,
 
 
 /* Encodes the bag into a SafeContents structure, and puts the output in
- * the given datum. Enc is set to non (0) if the data are encrypted;
+ * the given datum. Enc is set to non-zero if the data are encrypted;
  */
 int
 _pkcs12_encode_safe_contents (gnutls_pkcs12_bag_t bag, ASN1_TYPE * contents,
@@ -1288,7 +1326,7 @@ _pkcs12_encode_safe_contents (gnutls_pkcs12_bag_t bag, ASN1_TYPE * contents,
               goto cleanup;
             }
 
-          result = _gnutls_x509_write_value (c2, "?LAST.bagValue", &tmp, 0);
+          result = _gnutls_x509_write_value (c2, "?LAST.bagValue", &tmp);
 
           _gnutls_free_datum (&tmp);
 
@@ -1297,7 +1335,7 @@ _pkcs12_encode_safe_contents (gnutls_pkcs12_bag_t bag, ASN1_TYPE * contents,
         {
 
           result = _gnutls_x509_write_value (c2, "?LAST.bagValue",
-                                             &bag->element[i].data, 0);
+                                             &bag->element[i].data);
         }
 
       if (result < 0)
@@ -1344,7 +1382,7 @@ unsigned int i;
               gnutls_x509_crt_check_issuer((*extra_certs)[i], (*extra_certs)[i]) != 0)
             goto skip;
 
-          *chain = gnutls_realloc (*chain, sizeof((*chain)[0]) *
+          *chain = gnutls_realloc_fast (*chain, sizeof((*chain)[0]) *
                                                      ++(*chain_len));
           if (*chain == NULL)
             {
@@ -1368,37 +1406,37 @@ skip:
 
 /**
  * gnutls_pkcs12_simple_parse:
- * @p12: the PKCS#12 blob.
- * @password: optional password used to decrypt PKCS#12 blob, bags and keys.
+ * @p12: the PKCS12 blob.
+ * @password: optional password used to decrypt PKCS12 blob, bags and keys.
  * @key: a structure to store the parsed private key.
- * @chain: the corresponding to key certificate chain
- * @chain_len: will be updated with the number of additional
+ * @chain: the corresponding to key certificate chain (may be %NULL)
+ * @chain_len: will be updated with the number of additional (may be %NULL)
  * @extra_certs: optional pointer to receive an array of additional
- *                   certificates found in the PKCS#12 blob.
+ *               certificates found in the PKCS12 blob (may be %NULL).
  * @extra_certs_len: will be updated with the number of additional
- *                       certs.
- * @crl: an optional structure to store the parsed CRL.
+ *                   certs (may be %NULL).
+ * @crl: an optional structure to store the parsed CRL (may be %NULL).
  * @flags: should be zero or one of GNUTLS_PKCS12_SP_*
  *
- * This function parses a PKCS#12 blob in @p12blob and extracts the
+ * This function parses a PKCS12 blob in @p12blob and extracts the
  * private key, the corresponding certificate chain, and any additional
  * certificates and a CRL.
  *
- * The @extra_certs_ret and @extra_certs_ret_len parameters are optional
+ * The @extra_certs_ret and @extra_certs_len parameters are optional
  * and both may be set to %NULL. If either is non-%NULL, then both must
- * be.
+ * be set.
  * 
- * Encrypted PKCS#12 bags and PKCS#8 private keys are supported.  However,
+ * Encrypted PKCS12 bags and PKCS8 private keys are supported.  However,
  * only password based security, and the same password for all
  * operations, are supported.
  *
- * PKCS#12 file may contain many keys and/or certificates, and there
+ * A PKCS12 file may contain many keys and/or certificates, and there
  * is no way to identify which key/certificate pair you want.  You
- * should make sure the PKCS#12 file only contain one key/certificate
+ * should make sure the PKCS12 file only contain one key/certificate
  * pair and/or one CRL.
  *
- * It is believed that the limitations of this function is acceptable
- * for most usage, and that any more flexibility would introduce
+ * It is believed that the limitations of this function are acceptable
+ * for common usage, and that any more flexibility would introduce
  * complexity that would make it harder to use this functionality at
  * all.
  *
@@ -1546,7 +1584,6 @@ gnutls_pkcs12_simple_parse (gnutls_pkcs12_t p12,
               if (ret < 0)
                 {
                   gnutls_assert ();
-                  gnutls_x509_privkey_deinit (*key);
                   goto done;
                 }
 
@@ -1557,7 +1594,6 @@ gnutls_pkcs12_simple_parse (gnutls_pkcs12_t p12,
               if (ret < 0)
                 {
                   gnutls_assert ();
-                  gnutls_x509_privkey_deinit (*key);
                   goto done;
                 }
 
@@ -1685,7 +1721,7 @@ gnutls_pkcs12_simple_parse (gnutls_pkcs12_t p12,
                 { /* they don't match - skip the certificate */
                   if (extra_certs)
                     {
-                      _extra_certs = gnutls_realloc (_extra_certs,
+                      _extra_certs = gnutls_realloc_fast (_extra_certs,
                                                      sizeof(_extra_certs[0]) *
                                                      ++_extra_certs_len);
                       if (!_extra_certs)
@@ -1704,7 +1740,7 @@ gnutls_pkcs12_simple_parse (gnutls_pkcs12_t p12,
                 }
               else
                 {
-                  if (_chain_len == 0)
+                  if (chain && _chain_len == 0)
                     {
                       _chain = gnutls_malloc (sizeof(_chain[0]) * (++_chain_len));
                       if (!_chain)
@@ -1759,17 +1795,20 @@ gnutls_pkcs12_simple_parse (gnutls_pkcs12_t p12,
       gnutls_pkcs12_bag_deinit (bag);
     }
 
-  if (_chain_len != 1)
+  if (chain != NULL)
     {
-      ret = GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE;
-      goto done;
-    }
+      if (_chain_len != 1)
+        {
+          ret = GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE;
+          goto done;
+        }
 
-  ret = make_chain(&_chain, &_chain_len, &_extra_certs, &_extra_certs_len, flags);
-  if (ret < 0)
-    {
-      gnutls_assert();
-      goto done;
+      ret = make_chain(&_chain, &_chain_len, &_extra_certs, &_extra_certs_len, flags);
+      if (ret < 0)
+        {
+          gnutls_assert();
+          goto done;
+        }
     }
 
   ret = 0;
@@ -1815,8 +1854,11 @@ done:
       gnutls_free(_extra_certs);
     }
       
-  *chain = _chain;
-  *chain_len = _chain_len;
+  if (chain != NULL)
+    {
+      *chain = _chain;
+      *chain_len = _chain_len;
+    }
 
   return ret;
 }
