@@ -7,7 +7,7 @@
  *
  * The GnuTLS is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 3 of
+ * as published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful, but
@@ -825,6 +825,7 @@ read_pkcs_schema_params (schema_id * schema, const char *password,
     case PKCS12_3DES_SHA1:
     case PKCS12_ARCFOUR_SHA1:
     case PKCS12_RC2_40_SHA1:
+      memset(enc_params, 0, sizeof(*enc_params));
 
       if ((*schema) == PKCS12_3DES_SHA1)
         {
@@ -872,7 +873,7 @@ read_pkcs_schema_params (schema_id * schema, const char *password,
       if (enc_params->iv_size)
         {
           result =
-            _gnutls_pkcs12_string_to_key (2 /*IV*/, kdf_params->salt,
+            _gnutls_pkcs12_string_to_key (GNUTLS_MAC_SHA1, 2 /*IV*/, kdf_params->salt,
                                           kdf_params->salt_size,
                                           kdf_params->iter_count, password,
                                           enc_params->iv_size,
@@ -1766,7 +1767,7 @@ decrypt_data (schema_id schema, ASN1_TYPE pkcs8_asn,
       break;
     default:
       result =
-        _gnutls_pkcs12_string_to_key (1 /*KEY*/, kdf_params->salt,
+        _gnutls_pkcs12_string_to_key (GNUTLS_MAC_SHA1, 1 /*KEY*/, kdf_params->salt,
                                       kdf_params->salt_size,
                                       kdf_params->iter_count, password,
                                       key_size, key);
@@ -2112,7 +2113,7 @@ generate_key (schema_id schema,
 
     default:
       ret =
-        _gnutls_pkcs12_string_to_key (1 /*KEY*/, kdf_params->salt,
+        _gnutls_pkcs12_string_to_key (GNUTLS_MAC_SHA1, 1 /*KEY*/, kdf_params->salt,
                                       kdf_params->salt_size,
                                       kdf_params->iter_count, password,
                                       kdf_params->key_size, key->data);
@@ -2127,7 +2128,7 @@ generate_key (schema_id schema,
       if (enc_params->iv_size)
         {
           ret =
-            _gnutls_pkcs12_string_to_key (2 /*IV*/, kdf_params->salt,
+            _gnutls_pkcs12_string_to_key (GNUTLS_MAC_SHA1, 2 /*IV*/, kdf_params->salt,
                                           kdf_params->salt_size,
                                           kdf_params->iter_count, password,
                                           enc_params->iv_size,
