@@ -55,6 +55,10 @@
 #include "sockets.h"
 #include "benchmark.h"
 
+#ifdef HAVE_DANE
+#include <gnutls/dane.h>
+#endif
+
 #include <common.h>
 #include <socket.h>
 
@@ -64,7 +68,7 @@
 #define MAX_BUF 4096
 
 /* global stuff here */
-int resume, starttls, insecure, rehandshake, udp, mtu;
+int resume, starttls, insecure, ranges, rehandshake, udp, mtu;
 const char *hostname = NULL;
 const char *service = NULL;
 int record_max_size;
@@ -84,6 +88,8 @@ const char *x509_cafile = NULL;
 const char *x509_crlfile = NULL;
 static int x509ctype;
 static int disable_extensions;
+static int disable_sni;
+static unsigned int init_flags = GNUTLS_CLIENT;
 static const char * priorities = NULL;
 
 const char *psk_username = NULL;
@@ -114,7 +120,7 @@ static gnutls_privkey_t pgp_key = NULL;
 static void
 get_keyid (gnutls_openpgp_keyid_t keyid, const char *str)
 {
-  size_t keyid_size = sizeof (keyid);
+  size_t keyid_size = GNUTLS_OPENPGP_KEYID_SIZE;
 
   if (strlen (str) != 16)
     {
@@ -387,32 +393,60 @@ struct servent * sr;
   return sr->s_name;
 }
 
+static int service_to_port(const char* service)
+{
+unsigned int port;
+struct servent * sr;
+  
+  port = atoi(service);
+  if (port != 0) return port;
+
+  sr = getservbyname(service, udp?"udp":"tcp");
+  if (sr == NULL)
+    {
+      fprintf(stderr, "Warning: getservbyname() failed.\n");
+      exit(1);
+    }
+
+  return ntohs(sr->s_port);
+}
+
+
 static int
 cert_verify_callback (gnutls_session_t session)
 {
   int rc;
   unsigned int status = 0;
   int ssh = ENABLED_OPT(TOFU);
+#ifdef HAVE_DANE
+  int dane = ENABLED_OPT(DANE);
+#endif
+  int ca_verify = ENABLED_OPT(CA_VERIFICATION);
   const char* txt_service;
 
-  rc = cert_verify(session, hostname);
-  if (rc == 0)
+  print_cert_info (session, verbose, print_cert);
+
+  if (ca_verify)
     {
-      printf ("*** Verifying server certificate failed...\n");
-      if (!insecure && !ssh)
-        return -1;
-    }
-  else if (ENABLED_OPT(OCSP))
-    { /* off-line verification succeeded. Try OCSP */
-      rc = cert_verify_ocsp(session);
+      rc = cert_verify(session, hostname);
       if (rc == 0)
         {
-          printf ("*** Verifying (with OCSP) server certificate failed...\n");
+          printf ("*** Verifying server certificate failed...\n");
           if (!insecure && !ssh)
             return -1;
         }
-      else if (rc == -1)
-        printf("*** OCSP response ignored\n");
+      else if (ENABLED_OPT(OCSP) && gnutls_ocsp_status_request_is_checked(session, 0) == 0)
+        { /* off-line verification succeeded. Try OCSP */
+          rc = cert_verify_ocsp(session);
+          if (rc == 0)
+            {
+              printf ("*** Verifying (with OCSP) server certificate failed...\n");
+              if (!insecure && !ssh)
+                return -1;
+            }
+          else if (rc == -1)
+            printf("*** OCSP response ignored\n");
+        }
     }
 
   if (ssh) /* try ssh auth */
@@ -469,6 +503,43 @@ cert_verify_callback (gnutls_session_t session)
         }
     }
 
+#ifdef HAVE_DANE
+  if (dane) /* try DANE auth */
+    {
+      int port;
+      unsigned int sflags = ENABLED_OPT(LOCAL_DNS)?0:DANE_F_IGNORE_LOCAL_RESOLVER;
+      
+      port = service_to_port(service);
+      rc = dane_verify_session_crt( NULL, session, hostname, udp?"udp":"tcp", port, 
+                                    sflags, 0, &status);
+      if (rc < 0)
+        {
+          fprintf(stderr, "*** DANE verification error: %s\n", dane_strerror(rc));
+          if (!insecure)
+            return -1;
+        }
+      else
+        {
+          gnutls_datum_t out;
+
+          rc = dane_verification_status_print( status, &out, 0);
+          if (rc < 0)
+            {
+              fprintf(stderr, "*** DANE error: %s\n", dane_strerror(rc));
+              if (!insecure)
+                return -1;
+            }
+          
+          fprintf(stderr, "- DANE: %s\n", out.data);
+          gnutls_free(out.data);
+
+          if (status != 0 && !insecure)
+            return -1;
+        }
+
+    }
+#endif
+
   return 0;
 }
 
@@ -568,12 +639,12 @@ init_tls_session (const char *hostname)
   
   if (udp)
     {
-      gnutls_init (&session, GNUTLS_CLIENT|GNUTLS_DATAGRAM);
+      gnutls_init (&session, GNUTLS_DATAGRAM|init_flags);
       if (mtu)
         gnutls_dtls_set_mtu(session, mtu);
     }
   else
-    gnutls_init (&session, GNUTLS_CLIENT);
+    gnutls_init (&session, init_flags);
 
   if ((ret = gnutls_priority_set_direct (session, priorities, &err)) < 0)
     {
@@ -585,7 +656,7 @@ init_tls_session (const char *hostname)
 
   /* allow the use of private ciphersuites.
    */
-  if (disable_extensions == 0)
+  if (disable_extensions == 0 && disable_sni == 0)
     {
       if (!isdigit(hostname[0]) && strchr(hostname, ':') == 0)
         gnutls_server_name_set (session, GNUTLS_NAME_DNS, hostname,
@@ -604,7 +675,6 @@ init_tls_session (const char *hostname)
 
   gnutls_certificate_set_retrieve_function2 (xcred, cert_callback);
   gnutls_certificate_set_verify_function (xcred, cert_verify_callback);
-  gnutls_certificate_set_verify_flags (xcred, 0);
 
   /* send the fingerprint */
 #ifdef ENABLE_OPENPGP
@@ -625,9 +695,18 @@ init_tls_session (const char *hostname)
         }
     }
 
-#ifdef ENABLE_SESSION_TICKET
-  if (disable_extensions == 0 && !HAVE_OPT(NOTICKET)t)
-    gnutls_session_ticket_enable_client (session);
+  if (HAVE_OPT(HEARTBEAT))
+    gnutls_heartbeat_enable (session, GNUTLS_HB_PEER_ALLOWED_TO_SEND);
+
+#ifdef ENABLE_DTLS_SRTP
+  if (HAVE_OPT(SRTP_PROFILES))
+    {
+      ret = gnutls_srtp_set_profile_direct (session, OPT_ARG(SRTP_PROFILES), &err);
+      if (ret == GNUTLS_E_INVALID_REQUEST) fprintf (stderr, "Syntax error at: %s\n", err);
+      else 
+        fprintf(stderr, "Error in profiles: %s\n", gnutls_strerror(ret));
+      exit (1);
+    }
 #endif
 
   return session;
@@ -762,6 +841,9 @@ main (int argc, char **argv)
   int user_term = 0, retval = 0;
   socket_st hd;
   ssize_t bytes;
+#ifndef _WIN32
+  struct sigaction new_action, old_action;
+#endif
 
   set_program_name (argv[0]);
   cmd_parser (argc, argv);
@@ -775,10 +857,6 @@ main (int argc, char **argv)
       exit (1);
     }
 
-#ifdef ENABLE_PKCS11
-//  pkcs11_common ();
-#endif
-
   if (hostname == NULL)
     {
       fprintf (stderr, "No hostname given\n");
@@ -790,7 +868,6 @@ main (int argc, char **argv)
   init_global_tls_stuff ();
 
   socket_open (&hd, hostname, service, udp);
-  socket_connect (&hd);
 
   hd.session = init_tls_session (hostname);
   if (starttls)
@@ -844,7 +921,6 @@ main (int argc, char **argv)
           printf
             ("\n\n- Connecting again- trying to resume previous session\n");
           socket_open (&hd, hostname, service, udp);
-          socket_connect (&hd);
         }
       else
         {
@@ -876,7 +952,11 @@ after_handshake:
     }
 
 #ifndef _WIN32
-  signal (SIGALRM, &starttls_alarm);
+  new_action.sa_handler = starttls_alarm;
+  sigemptyset (&new_action.sa_mask);
+  new_action.sa_flags = 0;
+
+  sigaction (SIGALRM, &new_action, NULL);
 #endif
 
   fflush (stdout);
@@ -942,7 +1022,7 @@ after_handshake:
         }
 
       if (inp == IN_KEYBOARD)
-        {
+      {
           if ((bytes = read (fileno (stdin), buffer, MAX_BUF - 1)) <= 0)
             {
               if (hd.secure == 0)
@@ -980,7 +1060,17 @@ after_handshake:
                 }
             }
 
-          ret = socket_send (&hd, buffer, bytes);
+          if (ranges && gnutls_record_can_use_length_hiding(hd.session)) 
+            {
+	      gnutls_range_st range;
+	      range.low = 0;
+	      range.high = MAX_BUF;
+              ret = socket_send_range (&hd, buffer, bytes, &range);
+            }
+          else 
+            {
+        	  ret = socket_send(&hd, buffer, bytes);
+            }
 
           if (ret > 0)
             {
@@ -1070,11 +1160,16 @@ const char* rest = NULL;
       exit(0);
     }
 
+  disable_sni = HAVE_OPT(DISABLE_SNI);
   disable_extensions = HAVE_OPT( DISABLE_EXTENSIONS);
+  if (disable_extensions)
+    init_flags |= GNUTLS_NO_EXTENSIONS;
+  
   starttls = HAVE_OPT(STARTTLS);
   resume = HAVE_OPT(RESUME);
   rehandshake = HAVE_OPT(REHANDSHAKE);
   insecure = HAVE_OPT(INSECURE);
+  ranges   = HAVE_OPT(RANGES);
 
   udp = HAVE_OPT(UDP);
   mtu = OPT_VALUE_MTU;
@@ -1089,6 +1184,7 @@ const char* rest = NULL;
     }
 
   record_max_size = OPT_VALUE_RECORDSIZE;
+  
   fingerprint = HAVE_OPT(FINGERPRINT);
 
   if (HAVE_OPT(X509FMTDER))
@@ -1178,9 +1274,7 @@ do_handshake (socket_st * socket)
 {
   int ret;
 
-  gnutls_transport_set_ptr (socket->session,
-                            (gnutls_transport_ptr_t)
-                            gl_fd_to_handle (socket->fd));
+  gnutls_transport_set_int (socket->session, socket->fd);
   do
     {
       gnutls_handshake_set_timeout( socket->session, GNUTLS_DEFAULT_HANDSHAKE_TIMEOUT);
@@ -1196,12 +1290,11 @@ do_handshake (socket_st * socket)
   if (ret == 0)
     {
       /* print some information */
-      print_info (socket->session, print_cert, verbose);
+      print_info (socket->session, verbose, 0);
       socket->secure = 1;
     }
   else
     {
-      print_cert_info (socket->session, verbose, print_cert);
       gnutls_alert_send_appropriate (socket->session, ret);
       shutdown (socket->fd, SHUT_RDWR);
     }
@@ -1467,7 +1560,7 @@ cert_verify_ocsp (gnutls_session_t session)
       ret = -1;
       goto cleanup;
     }
-    
+
   ret = send_ocsp_request(NULL, crt, issuer, &resp, 1);
   if (ret < 0)
     {
@@ -1477,7 +1570,7 @@ cert_verify_ocsp (gnutls_session_t session)
     }
 
   /* verify and check the response for revoked cert */
-  ret = check_ocsp_response(issuer, &resp);
+  ret = check_ocsp_response(crt, issuer, &resp);
 
 cleanup:
   if (deinit_issuer)
